cpdef double[:,:] transpose(double[:,:] A)
cpdef double[:,:] matmult(double[:,:] A, double[:,:] B)
cpdef double[:] sub1d(double[:] arr1,double[:] arr2)
cpdef double[:] add1d(double[:] arr1,double[:] arr2)
cpdef double[:] mult1d(double[:] arr1,double[:] arr2)
cpdef double[:] pow1d(double[:] arr,double power)
cpdef double[:] fact1d(double[:] arr,double factor)
cpdef double[:,:] sub2d(double[:,:] arr1,double[:,:] arr2)
cpdef double[:,:] add2d(double[:,:] arr1,double[:,:] arr2)
cpdef double[:,:] mult2d(double[:,:] arr1,double[:,:] arr2)
cpdef double[:,:] pow2d(double[:,:] arr,double power)
cpdef double[:,:] fact2d(double[:,:] arr,double factor)
cpdef double[:,:,:] sub3d(double[:,:,:] arr1,double[:,:,:] arr2)
cpdef double[:,:,:] add3d(double[:,:,:] arr1,double[:,:,:] arr2)
cpdef double[:,:,:] mult3d(double[:,:,:] arr1,double[:,:,:] arr2)
cpdef double[:,:,:] pow3d(double[:,:,:] arr,double power)
cpdef double[:,:,:] fact3d(double[:,:,:] arr,double factor)
cpdef double[:] ravel2d(double[:,:] datain)
cpdef double[:] ravel3d(double[:,:,:] datain)
cpdef double[:,:] reshape2d(double[:,:] datain, Py_ssize_t[:] shapeout)
cpdef double[:,:,:] reshape3d(double[:,:,:] datain, Py_ssize_t[:] shapeout)
cdef double linearinterp(double[:,:] data, double x, double y)
cdef double[:] alinearinterp(double[:,:] data, double[:] x, double[:] y)
cpdef double nanmean(double[:] data)
cpdef tuple nonnanindexes(double[:] data)
cpdef double[:] nonnanarray(double[:] data)
cpdef double nanmedian(double[:] data)
cpdef double mad(double[:] data, int ignore_nan, int zero)
cpdef double[:] mad2d(double[:,:] data, int axis, int ignore_nan, int zero)
cpdef double[:,:] mad3d(double[:,:,:] data, int axis, int ignore_nan, int zero)
cpdef double quadratic_bisector(double[:] x, double[:] y)
cpdef long[:] arange(int start, int stop, int step)
cpdef double[:] linspace(double start, double stop, int num)
cdef list intmeshgrid(long[:] x, long[:] y)
cdef list meshgrid(double[:] x, double[:] y)
cdef double[:] aclip(double[:] val,double minval,double maxval)
cdef double clip(double val,double minval,double maxval)
cpdef Py_ssize_t[:] starbbox(double[:] coords,Py_ssize_t[:] imshape,double radius)
cpdef double[:,:] slice2d(double[:,:] array,Py_ssize_t[:] slc)
cpdef double[:,:] inverse(double[:,:] a)
cpdef double[:] qr_jac_solve(double[:,:] jac,double[:] resid,double[:] weight)
cpdef double[:,:] jac_covariance(double[:,:] jac,double[:] resid, double[:] weight)
cpdef double[:] poly2d(double[:,:] xdata, double[:] pars)
cpdef list jacpoly2d(double[:,:] xdata, double[:] pars)
cpdef list poly2dfit(double[:] x,double[:] y,double[:] data,double[:] error,int maxiter,double minpercdiff)
cpdef list relcoord(double[:] x, double[:] y, Py_ssize_t[:] shape)